import os
import glm
import numpy
import errno

from io import TextIOWrapper
from SpatialTransform import Euler

from .bvh import *
from .hierarchy import *
from .shared import *

# from shared.Pose import Pose


def parseLine(file: TextIOWrapper, lineNumber: int) -> tuple[int, list[str], tuple[TextIOWrapper, int, int, str]]:
    lineNumber += 1
    line = file.readline()
    tokens = line.strip().split()
    debugInfo = (file, lineNumber, len(line) - len(line.lstrip()) + len(tokens[0]), line)
    return (lineNumber, tokens, debugInfo)


def readAsBVH(path: str) -> BvhContainer:
    """Reads .bvh file as simple structure."""
    if not os.path.exists(path):
        raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), path)

    bvh = BvhContainer()
    with open(path, "r") as file:
        # check for 'HIERARCHY' start
        line, tokens, debugInfo = parseLine(file, 0)
        if not tokens[0] == 'HIERARCHY' and len(tokens) == 1:
            raise SyntaxError('First line must be only "HIERARCHY"', debugInfo)

        # parse joints recursivly
        line, tokens, debugInfo = parseLine(file, line)
        if tokens[0] != 'ROOT':
            raise SyntaxError('First Joint must be defined as "ROOT"', debugInfo)
        bvh.Root = parseJoint(file, deserializeJointName(tokens, debugInfo))

        # check for 'MOTION' start
        line, tokens, debugInfo = parseLine(file, line)
        if not tokens[0] == 'MOTION' or not len(tokens) == 1:
            raise SyntaxError('After end of hierarchy must follow "MOTION"', debugInfo)

        # check for frame count
        line, tokens, debugInfo = parseLine(file, line)
        if not tokens[0] == 'Frames:' or not len(tokens) == 2:
            raise SyntaxError('First line of "MOTION" section has to be "Frames: X"', debugInfo)
        else:
            bvh.FrameCount = deserializeFrameCount(tokens[1:], debugInfo)

        # check for frame rate
        line, tokens, debugInfo = parseLine(file, line)
        if not tokens[0] == 'Frame' or not len(tokens) == 3:
            raise SyntaxError('After frame count must follow "Frame Time"', debugInfo)
        else:
            bvh.FrameTime = deserializeFrameTime(tokens[2:], debugInfo)

        # parse motion data
        for _ in range(bvh.FrameCount):
            line, tokens, debugInfo = parseLine(file, line)
            deserializeMotion(bvh.Root, deserializeKeyframe(tokens, debugInfo))

        return bvh


def convertBvhToJoint(pose: RootPose) -> Joint:
    """Convers a simple bvh srtucture to a transform hierarchy."""
    joint = Joint(pose.Name, pose.Offset, pose.getRotation())
    joint.Keyframes = [Pose(f.Position, f.Rotation * joint.RotationLocal) for f in pose.Keyframes]

    for child in pose.Children:
        child = convertBvhToJoint(child)
        for frame in child.Keyframes:
            frame.Position = glm.inverse(joint.RotationLocal) * frame.Position
            frame.Rotation = glm.inverse(joint.RotationLocal) * frame.Rotation
        joint.attach(child, keepPosition=False, keepRotation=False, keepScale=False)
    return joint


def read(path: str) -> Joint:
    """Reads .bvh file as transform hierarchy."""
    return convertBvhToJoint(readAsBVH(path).Root).readPose(0, recursive=True)


def parseJoint(file: TextIOWrapper, name: str, line: int = 0) -> RootPose:
    # check for open bracket
    line, tokens, debugInfo = parseLine(file, line)
    if tokens[0] != '{' or len(tokens) > 1:
        raise SyntaxError('Joint definition must start with an opening bracket', debugInfo)

    # create joint and read channels and offeset
    joint = RootPose(name)
    joint.Offset = deserializeOffset(file, line)
    joint.Channels = deserializeChannles(file, line)

    # check for definition end or child joints or end site info
    while (True):
        line, tokens, debugInfo = parseLine(file, line)
        if tokens[0] == 'JOINT':
            joint.Children.append(parseJoint(file, deserializeJointName(tokens, debugInfo), line))
        elif tokens[0] == 'End':
            joint.EndSite = deserializeEndSite(file, line)
        elif tokens[0] == '}':
            break
        else:
            raise SyntaxError('Joint definition must end with an child joint, end site or closing bracket', debugInfo)

    # return deserialized joint root pose
    return joint


def deserializeJointName(tokens: list[str], debugInfo: tuple[TextIOWrapper, int, int, str]) -> str:
    if not isinstance(tokens, list) and len(tokens) != 2:
        raise SyntaxError('Joint header must be 2-dimensional tuple', debugInfo)
    return tokens[1]


def deserializeOffset(file: TextIOWrapper, line: int) -> glm.vec3:
    line, tokens, debugInfo = parseLine(file, line)
    if tokens[0] != 'OFFSET':
        raise SyntaxError('Expected OFFSET definition for joint', debugInfo)
    if not isinstance(tokens, list) or len(tokens[1:]) != 3:
        raise SyntaxError('Offset must be a 3-part tuple', debugInfo)
    try:
        return glm.vec3(list(map(float, tokens[1:])))
    except ValueError:
        raise SyntaxError('Offset must be numerics only', debugInfo)


def deserializeChannles(file: TextIOWrapper, line: int) -> list[str]:
    line, tokens, debugInfo = parseLine(file, line)
    if tokens[0] != 'CHANNELS':
        raise SyntaxError('Expected CHANNELS definition for joint', debugInfo)
    if not isinstance(tokens, list) or len(tokens) < 2:
        raise SyntaxError('Channels must be at least a 2-part tuple', debugInfo)
    try:
        channelCount = int(tokens[1])
    except ValueError:
        raise SyntaxError('Channel count must be numerical', debugInfo)
    if channelCount != len(tokens[2:]):
        raise SyntaxError('Channel count mismatch with labels', debugInfo)
    return tokens[2:]


def deserializeEndSite(file: TextIOWrapper, line: int) -> glm.vec3:
    line, tokens, debugInfo = parseLine(file, line)
    if tokens[0] != '{':
        raise SyntaxError('End Site definition must start with an opening bracket', debugInfo)
    result = deserializeOffset(file, line)
    line, tokens, debugInfo = parseLine(file, line)
    if tokens[0] != '}':
        raise SyntaxError('End Site definition must start with an opening bracket', debugInfo)
    return result


def deserializeFrameTime(data: list, debugInfo: tuple) -> float:
    if not isinstance(data, list) or len(data) != 1:
        raise SyntaxError('Frame time must be a 1-dimensional tuple', debugInfo)
    try:
        return float(data[0])
    except ValueError:
        raise SyntaxError('Frame time be numerical', debugInfo)


def deserializeFrameCount(data: list, debugInfo: tuple) -> float:
    if not isinstance(data, list) or len(data) != 1:
        raise SyntaxError('Frame count must be a 1-dimensional tuple', debugInfo)
    try:
        return int(data[0])
    except ValueError:
        raise SyntaxError('Frame time be numerical', debugInfo)


def deserializeKeyframe(data: list, debugInfo: tuple) -> numpy.ndarray:
    try:
        return numpy.array(list(map(float, data)))
    except ValueError:
        raise SyntaxError('Keyframe must be numerics only', debugInfo)


def deserializeMotion(joint: RootPose, data: numpy.ndarray, index=0) -> int:
    position = glm.vec3(joint.Offset)
    rotation = glm.vec3(0)
    rotOrder = ''

    for channel in joint.Channels:
        if 'Xposition' == channel: position.x = data[index]
        elif 'Yposition' == channel: position.y = data[index]
        elif 'Zposition' == channel: position.z = data[index]
        elif 'Xrotation' == channel: rotation.x = data[index]; rotOrder += 'X'
        elif 'Yrotation' == channel: rotation.y = data[index]; rotOrder += 'Y'
        elif 'Zrotation' == channel: rotation.z = data[index]; rotOrder += 'Z'
        index += 1
    joint.Keyframes.append(Pose(position, Euler.toQuatFrom(glm.radians(rotation), rotOrder, False)))

    for child in joint.Children:
        index = deserializeMotion(child, data, index)
    return index



import glm


class Pose:
    """Data structure for holding pose data."""

    @property
    def Position(self) -> glm.vec3: return glm.vec3(self._Position)
    @Position.setter
    def Position(self, value: glm.vec3) -> None: self._Position = glm.vec3(value)

    @property
    def Rotation(self) -> glm.quat: return glm.quat(self._Rotation)
    @Rotation.setter
    def Rotation(self, value: glm.quat) -> None: self._Rotation = glm.quat(value)

    @property
    def Scale(self) -> glm.vec3: return glm.vec3(self._Scale)
    @Scale.setter
    def Scale(self, value: glm.vec3) -> None: self._Scale = glm.vec3(value)

    def __init__(self, position: glm.vec3 = glm.vec3(), rotation: glm.quat = glm.quat(), scale: glm.vec3 = glm.vec3(1)) -> None:
        self._Position = glm.vec3(position)
        self._Rotation = glm.quat(rotation)
        self._Scale = glm.vec3(scale)

    def __repr__(self) -> str:
        return (f"Pos: {self.Position}, Rot: {self.Rotation}, Scale: {self.Scale}")

    def __str__(self) -> str:
        return self.__repr__()



import glm
from SpatialTransform import Transform
from ..shared import Pose


class Joint(Transform):
    """Transform data of a joint including its keyframes.

    DataBVH holds the original bvh data of the hierarchy and motion.

    Keyframes holding the converted motion data, where position and rotation are in local space, in cluding the bvh base pose"""

    @property
    def Parent(self) -> "Joint": return self._Parent

    @property
    def Children(self) -> list["Joint"]: return self._Children

    @property
    def CurrentFrame(self) -> int: return self._CurrentFrame

    @property
    def Keyframes(self) -> list[Pose]: return self._Keyframes
    @Keyframes.setter
    def Keyframes(self, value: list[Pose]) -> None: self._Keyframes = list(value)

    def __init__(self, name: str, position: glm.vec3 = glm.vec3(), rotation: glm.quat = glm.quat(), scale: glm.vec3 = glm.vec3(1), emptyKeyframes: int = 0) -> None:
        super().__init__(name, position, rotation, scale)
        self._Parent: "Joint" = None
        self._Children: list["Joint"] = []

        if emptyKeyframes > 0:
            self._Keyframes: list[Pose] = [Pose(self._PositionLocal, self.RotationLocal, self.ScaleLocal) for _ in range(emptyKeyframes)]
            self.readPose(0)
        else:
            self._Keyframes: list[Pose] = []
            self._CurrentFrame = -1

    def readPose(self, frame: int, recursive: bool = True) -> "Joint":
        """Sets the transform position and rotation properties by the given keyframe.

        If recursive the children do also load their keyframe data."""
        if frame < 0: frame += len(self.Keyframes)
        if frame < 0 or frame >= len(self._Keyframes):
            raise ValueError(f'Frame number "{frame}" for {self.Name} is out of range. ({len(self._Keyframes)} Keyframes)')

        self._CurrentFrame = frame
        self.PositionLocal = self._Keyframes[frame].Position
        self.RotationLocal = self._Keyframes[frame].Rotation
        self.ScaleLocal = self._Keyframes[frame].Scale

        if recursive:
            for child in self.Children:
                child.readPose(frame, recursive)
        return self

    def writePose(self, frame: int, recursive: bool = True) -> "Joint":
        """Sets the keyframe properties by the current transform local position and local rotation.

        If recursive the children do also update their keyframe data."""
        if frame < 0: frame += len(self.Keyframes)
        if frame < 0 or frame > len(self._Keyframes):
            raise ValueError(f'Frame number "{frame}" for {self.Name} is out of range.')

        elif frame < len(self._Keyframes):
            self._Keyframes[frame].Position = self.PositionLocal
            self._Keyframes[frame].Rotation = self.RotationLocal
            self._Keyframes[frame].Scale = self.ScaleLocal

        elif frame == len(self._Keyframes):
            self._Keyframes.append(Pose(self.PositionLocal, self.RotationLocal))

        if recursive:
            for child in self.Children:
                child.writePose(frame)
        return self

    def attach(self, *nodes: "Joint", keepPosition: bool = False, keepRotation: bool = False, keepScale: bool = False) -> "Joint":
        for node in nodes:
            # attach the joint
            super().attach(node, keepPosition=keepPosition, keepRotation=keepRotation, keepScale=keepScale)

            # apply attach corrections to animation data
            for pose in node._Keyframes:
                if keepPosition: pose.Position = self.SpaceWorldInverse * pose.Position
                if keepRotation: pose.Rotation = self.RotationWorldInverse * pose.Rotation
                if keepScale: pose.Scale = self.ScaleWorldInverse * pose.Scale
        return self

    def detach(self, node: "Joint", keepPosition: bool = False, keepRotation: bool = False, keepScale: bool = False) -> "Joint":

        # apply detach corrections to animation data
        for pose in node._Keyframes:
            if keepPosition: pose.Position = self.SpaceWorld * pose.Position
            if keepRotation: pose.Rotation = self.RotationWorld * pose.Rotation
            if keepScale: pose.Scale = self.ScaleWorld * pose.Scale

        # finally detatch the joint
        return super().detach(node, keepPosition, keepRotation, keepScale)

    def clearParent(self, keepPosition: bool = False, keepRotation: bool = False, keepScale: bool = False) -> "Joint":
        return super().clearParent(keepPosition, keepRotation, keepScale)

    def clearChildren(self, keepPosition: bool = False, keepRotation: bool = False, keepScale: bool = False) -> "Joint":
        return super().clearChildren(keepPosition, keepRotation, keepScale)

    def applyPosition(self, position: glm.vec3 = None, recursive: bool = False) -> "Joint":
        # define positional change
        change = -self.PositionLocal if position is None else position
        changeInverse = glm.inverse(self.RotationLocal) * (glm.div(1, self.ScaleLocal) * -change)

        # apply to current data
        super().applyPosition(position, recursive=recursive)

        # apply to keyframes
        for pose in self._Keyframes:
            pose.Position += change
        for child in self._Children:
            for pose in child._Keyframes:
                pose.Position += changeInverse

        return self

    def applyRotation(self, rotation: glm.quat = None, recursive: bool = False) -> "Joint":
        # define rotational change
        change = glm.inverse(self.RotationLocal) if rotation is None else rotation
        changeInverse = glm.inverse(change)

        # apply to current data
        super().applyRotation(rotation, recursive=False)

        # apply to keyframes
        for pose in self._Keyframes:
            pose.Rotation = change * pose.Rotation
        for child in self._Children:
            for pose in child._Keyframes:
                pose.Position = changeInverse * pose.Position
                pose.Rotation = changeInverse * pose.Rotation
            if recursive:
                child.applyRotation(rotation, recursive=True)

        return self

    def appyScale(self, scale: glm.vec3 = None, recursive: bool = False) -> "Joint":
        # define change in scale
        changeFrame = glm.div(self.ScaleLocal, self._Keyframes[self._CurrentFrame].Scale)
        change = glm.div(1, self.ScaleLocal) if scale is None else scale
        changeInverse = glm.div(1, change)

        # apply to current data
        super().appyScale(scale, recursive=False)

        # apply to keyframes
        for pose in self._Keyframes:
            pose.Scale *= changeFrame * change
        for child in self._Children:
            for pose in child._Keyframes:
                pose.Position *= changeInverse
                pose.Scale *= changeInverse
            if recursive:
                child.appyScale(scale, recursive=True)

        return self

    def roll(self, degrees: float, recursive: bool = False) -> "Joint":
        """Rolls the joint along its local Y axis. Updates the children so there is no spatial change

        Returns itself"""
        change = glm.angleAxis(glm.radians(degrees), (0, 1, 0))
        changeInverse = glm.inverse(change)

        self.RotationLocal = self.RotationLocal * change
        for pose in self._Keyframes:
            pose.Rotation = pose.Rotation * change

        for child in self._Children:
            child.PositionLocal = changeInverse * child.PositionLocal
            child.RotationLocal = changeInverse * child.RotationLocal
            for pose in child._Keyframes:
                pose.Position = changeInverse * pose.Position
                pose.Rotation = changeInverse * pose.Rotation
            if recursive: child.roll(degrees, recursive=True)

        return self

    def filter(self, pattern: str, isEqual: bool = False, caseSensitive: bool = False) -> list["Joint"]:
        return super().filter(pattern, isEqual, caseSensitive)

    def filterRegex(self, pattern: str) -> list["Joint"]:
        return super().filterRegex(pattern)

    def layout(self, index: int = 0, depth: int = 0) -> list[tuple["Joint", int, int]]:
        return super().layout(index, depth)



from .RootPose import RootPose


class BvhContainer:
    """Container for the information of the bvh file.

    Root is the recursive definition of the skeleton.

    Frame time the frame time.

    Frams are the count of keyframes of the motion."""
    Root: RootPose
    FrameTime: float
    FrameCount: int

    def __init__(self):
        self.Root: RootPose = None
        self.FrameTime: float = None
        self.FrameCount: int = None
